#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
## This file may be used under the terms of the GNU General Public
## License version 2.0 as published by the Free Software Foundation
## and appearing in the file LICENSE included in the packaging of
## this file. 
##
## This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
## WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
##
## See the NOTICE file distributed with this work for additional
## information regarding copyright ownership.

import os, sys, shlex, subprocess
from xml.dom import minidom

class cpufreqInfo (object):
    """Author : Yves-Gwenael Bourhis
    
    requires cpufrequtils : http://www.kernel.org/pub/linux/utils/kernel/cpufreq/cpufrequtils.html"""
    
    def __init__(self, cpunum = None):
        """Create properties in camelCase format:
        
        the property from 'cpufreq-info --affected-cpus' will be in self.affectedCpus
        the property from 'cpufreq-info --governors' will be in self.governors
        etc.
        
        read "man cpufreq-info" for more information on cpufreq-info itself.
        
        For "easy remembering" I made the following correlation:
        properties of this class will return a tuple (because tuples cannot be modified)
        methods of this class will return a list (because lists can be modified)
        However, it's useless modifying a returned list in order to set a new state.
        """
        
        self._cpunum = cpunum
        #Create properties:
        #We call here "properties", values obtained by cpufreq-info,
        #which are hardware dependent,
        #and which are NOT subject to change at runtime.
        self._property_list = ['affected-cpus', 'governors', 'driver', 'hwlimits'] #'affected-cpus' HAS to be the first element
        if self._cpunum != None:
            self._property_list.pop(0)
        self._create_properties (self._property_list)
        
        #Create methods:
        #We call here "methods", values obtained by cpufreq-info,
        #which are subject to change at runtime,
        #Therefore we create a method to obtain the curent value.
        self._method_list = ['freq', 'hwfreq', 'policy'] #The order in this list has no importance, rearange as you want
        self._create_methods (self._method_list)
        
        #create for each cpu the same methods and properties as above:
        if self._cpunum == None:
            self.cpu = []
            for cpu in self.affectedCpus:
                self.cpu.append(cpufreqInfo(cpu))
        
        
    def _create_properties (self, property_list):
        """create object properties from the property_list in camelCase format (ex.: affected-cpus will be in self.affectedCpus )"""        
        for prop in property_list:
            prop = str(prop)
            propname = self._camelCase(prop)
            command_line = 'cpufreq-info --' + prop
            #vars(self)[propname] = self._run_command(command_line)
            setattr( self, propname, self._run_command(command_line) )
            
            
    def _create_methods(self, methods_list, cpunum=None):
        """create object methods from the methods_list in camelCase format (ex.: freq will be self.freq() )"""        
        for meth in methods_list:
            meth = str(meth)
            methname = self._camelCase(meth)
            func = lambda cpu=cpunum, value=meth : self._getavalue(value, cpu, 'list')
            #vars(self)[methname] = func
            setattr(self, methname, func)
            
            
    def _getavalue(self, whatvalue, cpunum=None, resulttype=None):
        command_line = 'cpufreq-info -c '+ str(cpunum) + ' --' + str(whatvalue)
        args = shlex.split (command_line)
        if cpunum == None:
            args.pop(1)
            args.pop(1)
        #print args
        return self._run_command(args, resulttype)

        
    def _run_command(self, command_line, resulttype=None):
        """returns a tupple containing the stdout output of the command"""        
        #is command_line a string or a list:
        if isinstance(command_line , basestring):
            args = shlex.split (command_line)
        elif isinstance(command_line , list):
            args = command_line
        else:
            return ('Bogus Command','Nothing to do')
        cmdOpen = subprocess.Popen(args, stdout=subprocess.PIPE)
        result = cmdOpen.communicate()[0]
        #cmdOpen.terminate()
        if resulttype in [None, 'tuple']:
            return tuple( result.strip().split() )
        elif resulttype == 'list':
            return list( result.strip().split() )
        elif resulttype == 'text':
            return result
        
    def _camelCase(self, word):
        """format the word argument to camelCase (ex.: if word ='affected-cpus' this method returns 'affectedCpus')"""
        wordsplit = word.split('-')
        camelWord = ''
        i = 0
        for part in wordsplit:
            if i > 0:
                wordsplit[i] = part.capitalize()
            camelWord += wordsplit[i]
            i += 1
        return camelWord
    
    
    def __str__(self):
        #Create DOM object
        DOMimpl = minidom.getDOMImplementation()

        #Create XML Document
        xmldoc = DOMimpl.createDocument(None, "cpufreqinfo", None)
        doc_root = xmldoc.documentElement
                
        #Add a General Info Node:
        if self._cpunum == None:
            command_line = 'cpufreq-info --human'
            node = xmldoc.createElement('general-info')
            propvalue = self._run_command(command_line, 'text')
            node.appendChild(xmldoc.createTextNode(propvalue))
            doc_root.appendChild(node)
        
        #Add Property Nodes
        propsandmeths = self._property_list + self._method_list
        for prop in propsandmeths:
            prop = str(prop)
            
            #Create Node
            node = xmldoc.createElement(prop)
            propvalues = self._getavalue(prop, self._cpunum) #Obtain list of values
            for propvalue in propvalues:
                node.appendChild(xmldoc.createTextNode(propvalue)) #Add values in node
            
            #Add Node
            doc_root.appendChild(node)
        #return xmldoc.toxml()
        return xmldoc.toprettyxml()
