#! /usr/bin/env python
# -*- coding: utf-8 -*-
#
## This file may be used under the terms of the GNU General Public
## License version 2.0 as published by the Free Software Foundation
## and appearing in the file LICENSE included in the packaging of
## this file. 
##
## This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
## WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
##
## See the NOTICE file distributed with this work for additional
## information regarding copyright ownership.

from dbus.exceptions import DBusException

import getopt
import os.path
import string
import subprocess
import sys
import tokenize
import traceback
import pwd

from signal import SIGKILL
from ConfigParser import SafeConfigParser
from util import util

import os.path
import pwd
import re
import shutil
import time
from datetime import datetime
import glob
import dbus

from util import rrd
from threading import Thread
import threading
from powertop import PowerTop
from cpufrequtils import cpufreqInfo

class Deskolo(object):
    
    def __init__(self, conf):
        self.config = SafeConfigParser()
        self.config_file = conf
        self.config.read(self.config_file)

        self.jobfile = self.config.get("deskolo", "jobfile")
        self.keyboardInput = self.config.get("deskolo", "keyboard")
        self.mouseInput = self.config.get("deskolo", "mouse")
        self.wattsupDevice = self.config.get("deskolo", "wattsup")
        self.intervals = eval(self.config.get("deskolo", "intervals"))
        self.powerTopInterval = self.config.get("deskolo", "powertop-duration")
        self.scriptDir = self.config.get("deskolo", "script-dir")
        
        self.logDir = "/var/log"
        self.logFile = "%s/%s" % (self.logDir, "deskolo.log")
        self.logger = open(self.logFile, "a+")
        
        self.dataDir = "/var/lib/deskolo"
        if not os.path.exists(self.dataDir):
            os.makedirs(self.dataDir)
        
        self.cpufreqInfoFlag = True

    def checkIsProcessRunning(self, name):
        cmd = "ps -eo args | grep -i %s" % name
        output = util.shell_exec(cmd)
        lines = output.split("\n")
        found = False
        for line in lines:
            if  line and line.find(name) == 0:
                found = True
                break
        return found  
        


    def runWattsup(self):
        measures = ["watts", "volts", "amps", "kwh", "duty-cycle", "power-cycle"]
        #enable debug message for filling the pipe buffer more rapidly, so that we don't wait long
        #before the wattsuplog.py starts processing some data (useful for debugging wattsuplog.py)
        #/usr/bin/wattsup -d --interval=1 -c 0 -r -t ttyUSB0 watts volts amps kwh duty-cycle power-cycle
        cmd = ["/usr/bin/wattsup", "-d", "--interval=%s" % self.intervals["wattsup"], "-c", "0", "-r", "-t", self.wattsupDevice]
        if not os.path.exists(cmd[0]):
            self.log("wattsup not found - the energy consumption won't be logged")
            return
        
        for measure in measures:
            cmd.append(measure)
        
        self.wattsup = subprocess.Popen(cmd, stderr=self.logger, stdout=subprocess.PIPE)

        #a delay is needed for the process to start indeed
        #if not self.checkIsProcessRunning("wattsup"):
        #self.log("wattsup was not launched successfully, please check that the device is plugged in correctly.")
        #return
        
        cmd = ["python", "%s/wattsuplog.py" % self.scriptDir]
        self.wattsuplog = subprocess.Popen(cmd, stdin=self.wattsup.stdout, stderr=self.logger, stdout=self.logger)
    
    def runPowerTop(self, epoch):
        if not os.path.exists("/usr/bin/powertop"):
            self.log("powertop not found")
            return

        pt = PowerTop(self.powerTopInterval, self.logger, self.scriptDir)
        pt.start()
        
    def doCpufreqInfo(self):
        if not os.path.exists("/usr/bin/cpufreq-info"):
            self.log("cpufreq-info not found - the cpufreq information won't be logged")
            return        
        
        if self.cpufreqInfoFlag:
            cfi = cpufreqInfo()
            path = "%s/xml/cpufreq/cpufreq-%s.xml" % ("/var/lib/deskolo", util.getFormattedDate())
        
            if not os.path.exists("%s/xml/cpufreq" % ("/var/lib/deskolo")):
                os.makedirs("%s/xml/cpufreq" % ("/var/lib/deskolo"))
            
            # write the string
            f = open(path, "w")
            f.write(cfi.__str__())
            f.close()
            self.cpufreqInfoFlag = False     
        

    def logCpu(self):
        #self.exportSarQuantity(["-P","ALL"])
        self.exportSarQuantity(["-u"], "cpu")
        
    def logMemory(self):
        self.exportSarQuantity(["-r"], "memory")

    def logSwap(self):
        self.exportSarQuantity(["-S"], "swap")

    def logNetwork(self):
        self.exportSarQuantity(["-n", "DEV"], "network")
        
    def logIo(self):
        self.exportSarQuantity(["-b"], "io")
    
    def logLoad(self):
        self.exportSarQuantity(["-q"], "load")
        
    def logKeyboard(self, epoch):
        
        #kill the previous keyboard logger process
        if hasattr(self, "keyboardMonitor"):
            try:
                keylog = os.path.getsize(self.keyboardLogger.name) / 96
                logQuantity(self.dataDir, "keyboard", "keylog", keylog, self.intervals["keyboard"], epoch)
                self.keyboardMonitor.terminate()
                
            except:
                traceback.print_exc(file=sys.stdout)
        
        cmd = ["cat", self.keyboardInput]
        self.keyboardLogger = open("%s/%s" % (self.dataDir, "keylog"), "w")
        self.keyboardMonitor = subprocess.Popen(cmd, stderr=self.logger, stdout=self.keyboardLogger)

    def logMouse(self, epoch):
        #kill the previous keyboard logger process
        if hasattr(self, "mouseMonitor"):
            try:
                mouselog = os.path.getsize(self.mouseLogger.name) / 96
                logQuantity(self.dataDir, "mouse", "mouselog", mouselog, self.intervals["mouse"], epoch)
                self.mouseMonitor.terminate()
                
            except:
                traceback.print_exc(file=sys.stdout)
            
        
        cmd = ["cat", self.mouseInput]
        self.mouseLogger = open("%s/%s" % (self.dataDir, "mouselog"), "w")
        self.mouseMonitor = subprocess.Popen(cmd, stderr=self.logger, stdout=self.mouseLogger)
        
        
    def exportSarQuantity(self, options, logprefix):
        now = datetime.now()
        cmd = ["sadf", "-d", "/var/log/sa/sa%s" % str(now.day), "--"]
        for option in options:
            cmd.append(option)
        #the file is overwritten each time
        logger = open("%s/%s-%s" % (self.dataDir, logprefix, util.getFormattedDate()), "w")
        sadf = subprocess.Popen(cmd, stdout=logger)
        sadf.wait()
        logger.flush()
        logger.close()
        
    
    def logBrightness(self, epoch):
        bus = dbus.SystemBus()
        try:
            obj = bus.get_object('org.freedesktop.Hal', '/org/freedesktop/Hal/devices/computer_backlight')
            iface = dbus.Interface(obj, 'org.freedesktop.Hal.Device.LaptopPanel')
            brightness = iface.GetBrightness()
            logQuantity(self.dataDir, "screen", "brightness", brightness, self.intervals["brightness"], epoch)
        except DBusException, e:
            self.log(str(e))

    def logScreenSaverState(self, epoch):
        cmd = ["/usr/bin/xdg-screensaver", "status"]
        if not os.path.exists(cmd[0]):
            self.log("xdg-screensaver not found - the screen state won't be logged")
            return
        xss = subprocess.Popen(cmd, stderr=self.logger, stdout=subprocess.PIPE)
        xsss = 0
        data = xss.communicate()[0]
        if data == "enabled":
            xsss = 1

        logQuantity(self.dataDir, "screen", "screensaver", xsss, self.intervals["screensaver"], epoch)

    def runDpmsMonitor(self):
        cmd = ["/usr/bin/xlib-dpms"]
        if not os.path.exists(cmd[0]):
            self.log("xlib-dpms not found - the DPMS level won't be logged")
            return

        self.dpms = subprocess.Popen(cmd, stderr=self.logger, stdout=subprocess.PIPE)
        
        #a few seconds are needed for the process to start
        #if not self.checkIsProcessRunning("xlib-dpms"):
        #    self.log("xlib-dpms was not launched successfully.")
        
        cmd = ["python", "%s/dpmslog.py" % self.scriptDir]
        self.dpmslog = subprocess.Popen(cmd, stdin=self.dpms.stdout, stderr=self.logger, stdout=self.logger)
    
    def logScreenSize(self, epoch):
        cmd = ["/usr/bin/xdpyinfo"]
        if not os.path.exists(cmd[0]):
            self.log("xdpyinfo not found - the screen size won't be logged")
            return
        xdpyinfo = subprocess.Popen(cmd, stderr=self.logger, stdout=subprocess.PIPE)
        data = xdpyinfo.communicate()[0].strip()
        re_dimensions = re.compile(r"(\s*)dimensions:(\s*)(.*)\spixels\s\((.*)\)\n")
        result = re_dimensions.search(data)
        if result:
            size = result.group(3)
            idx = size.find("x")
            width = int(size[0:idx])
            length = int(size[idx + 1:])
            size = width * length
        else:
            size = ""
            
        logQuantity(self.dataDir, "screen", "size", size, self.intervals["screensize"], epoch)
   
    def logTemperature(self):
        #the temperature is actually logged by collectd
        try:
            temperature = util.read_as_string("/sys/class/thermal/thermal_zone0/temp")
            temperature = temperature.strip()
        except Exception, e:
            self.log(str(e))
    
    def run(self):
        flag = True
        
        if not self.checkIsProcessRunning("wattsup"):
            self.log("launching wattsup")
            self.runWattsup()
        
        self.runDpmsMonitor()

        self.threads = []
        ar = {
                "brightness":"logBrightness",
                "screensize":"logScreenSize",
                "mouse":"logMouse",
                "keyboard":"logKeyboard",
                "powertop":"runPowerTop",
                "screensaver":"logScreenSaverState"
             }
            
        while flag:
            time.sleep(1)

            epoch = time.time()
            epochint = int(epoch)
            for key, value in ar.iteritems():
                if epochint % self.intervals[key] == 0:
                    try:
                        self.doCpufreqInfo()
                        self.log("%s [epoch: %s]" % (value, str(epoch)))
                        eval("self.%s(%d)" % (value, epochint))
                        #target = eval("self.%s" % value)
                        #self.controller.addJob(epochint, target)
                        #t = Thread(target=target, args=(epochint,))
                        #t.start()
                        #regarder dans la target s'il y a un thread déjà actif pour ce target:
                        #si c'est le cas: on attend que le thread termine avant de continuer
                        #sinon ok on lance un thread pour ce target
                        
                    except:
                        traceback.print_exc(file=sys.stdout)
                
            flag = os.path.exists(self.jobfile)
        self.terminate()

    def terminate(self):
        try:
            for thread in self.threads:
                thread.cancel()
        
            if hasattr(self, "wattsup"):
                self.wattsup.terminate()
            
            if hasattr(self, "dpms"):
                self.dpms.terminate()
            
            
            if hasattr(self, "keyboardWatcher"):
                self.keyboardMonitor.terminate()
                
            if hasattr(self, "mouselog"):
                self.mouseMonitor.terminate()
            
#            if hasattr(self, "wattsupLogger"):
#                self.wattsupLogger.flush()
#                self.wattsupLogger.close()

        except Exception, e:
            traceback.print_exc(sys.stdout)


    def stop(self):
        if os.path.exists(self.jobfile):
            #self.terminate()
            os.remove(self.jobfile)
        else:
            if self.checkIsRunning():
                self.terminate()
                output = util.shell_exec(self.config.get("deskolo", "pcheck"))
                for line in output.split("\n"):
                    if  line and line.find("deskolo.py start") >= 0:
                        data = line.split()
                        pid = int(data[1].strip())
                        try:
                            while 1:
                                os.kill(pid, SIGKILL)
                                time.sleep(0.1)
                                os.remove(self.jobfile)
                        except Exception, err:
                            err = str(err)
                            if not err.find("No such process") > 0:
                                print str(err)
                                sys.exit(1)
                        finally:
                            break
            else:
                print "No running Deskolo found"

    def checkIsRunning(self):
        output = util.shell_exec(self.config.get("deskolo", "pcheck"))
        lines = output.split("\n")
        foundOther = False
        counter = 0
        for line in lines:
            if  line and line.find("deskolo.py") >= 0:
                counter += 1
                if counter > 1:
                    foundOther = True
                    break  

        return foundOther

    def start(self):
        #tuples = output.split(" ")
        #if os.path.exists(pid_path) and read_as_string(pid_path) is not None:
        if  not self.checkIsRunning():
            self.log("starting deskolo")
            open(self.jobfile, 'w+')
            try:
                self.run()
            finally:
                self.terminate() 
        else:
            print "There is already a Deskolo daemon running, aborting."
            sys.exit(2)
            #process = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, close_fds=True)
            #output = process.stdout.read()
            #append_string(str(process.pid), pid_path)

    def status(self):
        flag = os.path.exists(self.jobfile)
        if flag:
            lmtime = os.path.getmtime(self.jobfile)
            print "Deskolo running since %s" % lmtime
        else:
            print "Deskolo not started"
        
    def log(self, message, level=0):
        if (level == 0):
            log_level = "INFO"
        elif (level == 1):
            log_level = "WARNING"
        else:
            log_level = "ERROR"

        now = datetime.now()
        now_str = datetime.strftime(now, "%Y-%m-%d %H:%M:%S")
        
        self.logger.write("[%s] [%s] %s\n" % (now_str, log_level, message))
        self.logger.flush()
        sys.stdout.write("%s [deskolo] %s\n" % (now_str, message))

def checkIsRoot():
    user = pwd.getpwuid(os.getuid())[0]
    if user != "root":
        print ("Deskolo must be executed as root. Exiting.")
        sys.exit(os.EX_NOPERM)    

def logQuantity(datadir, qtytype, qtyname, qtyvalue, interval, epoch):
        
    logQuantityCsv(datadir, epoch, qtytype, qtyname, qtyvalue)
    logQuantityRrd(datadir, epoch, qtytype, qtyname, qtyvalue, interval)

def logQuantityCsv(datadir, epoch, qtytype, qtyname, qtyvalue):
    csvdatadir = checkDataDir(datadir, "csv", qtytype)
    
    logger = open("%s/%s-%s" % (csvdatadir, qtyname, util.getFormattedDate()), "a+")    
    if os.path.getsize(logger.name) == 0:
        logger.write("#epoch,%s\n" % qtyname)
    
    logger.write("%s,%s\n" % (epoch, qtyvalue))
    logger.flush()
    logger.close()
        
def logQuantityRrd(datadir, epoch, qtytype, qtyname, qtyvalue, interval):
    if not os.path.exists("/usr/bin/rrdtool"):
        return
    
    rrddatadir = checkDataDir(datadir, "rrd", qtytype)

    path = "%s/%s.rrd" % (rrddatadir, qtyname)
    rrddb = rrd.RRD(path, vertical_label=qtyname,)
    
    if not os.path.exists(path):
        rrddb.createDatabase(interval)
    
    rrddb.update(epoch, qtyvalue)
        
def checkDataDir(datadir, format, qtytype):
    qtydatadir = "%s/%s/%s" % (datadir, format, qtytype)
    if not os.path.exists(qtydatadir):
        os.makedirs(qtydatadir)
    
    return qtydatadir

def usage():
    usage = "Usage: deskolo start|stop|status|setup"
    print usage

def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "")
    except getopt.GetoptError:
        usage()
        sys.exit(2)
    
    if len(args) < 1:
        usage()

    for arg in args:
        if arg in ["start", "stop", "restart", "status"]:
            try:
                checkIsRoot()
                deskolo = Deskolo("/etc/deskolo.cfg")
                eval("deskolo.%s()" % arg)
            except KeyboardInterrupt:
                sys.exit(0)
            except Exception, e:
                traceback.print_exc(file=sys.stdout)
                sys.exit(0)
        else:
            usage()


        
if __name__ == "__main__":
    main()
